import React from "react";
import Foundation from "@microsoft/fast-components-foundation-react";
import { isNil } from "lodash-es";
import { classNames } from "@microsoft/fast-web-utilities";
import { extractHtmlElement } from "@microsoft/fast-react-utilities";
import { DisplayNamePrefix } from "../utilities";
import { DataGridContext } from "./data-grid-context";
class DataGridCell extends Foundation {
    constructor() {
        super(...arguments);
        this.handledProps = {
            managedClasses: void 0,
            rowData: void 0,
            column: void 0,
            columnIndex: void 0,
        };
        this.focusTarget = React.createRef();
        this.rootElement = React.createRef();
        /**
         * returns true if this is the datagrid's current focus cell
         */
        this.isDesiredFocusCell = () => {
            if (this.context === null) {
                return false;
            }
            return (this.props.rowData[this.context.dataRowKey] ===
                this.context.desiredFocusRowKey &&
                this.props.column.columnDataKey === this.context.desiredFocusColumnKey);
        };
        /**
         * returns true if this is the datagrid's current focus cell
         */
        this.isFocusCell = () => {
            if (this.context === null) {
                return false;
            }
            return (this.props.rowData[this.context.dataRowKey] === this.context.focusRowKey &&
                this.props.column.columnDataKey === this.context.focusColumnKey);
        };
        /**
         * Handle the keydown event of the item
         */
        this.handleKeyDown = (e) => {
            this.context.onCellKeyDown(this.props, e);
        };
        /**
         * Handle focus event
         */
        this.handleFocus = (e) => {
            const focusTargetElement = extractHtmlElement(this.focusTarget);
            if (focusTargetElement !== null && focusTargetElement instanceof HTMLElement) {
                focusTargetElement.focus();
            }
            this.context.onCellFocused(this.props, e);
        };
    }
    /**
     * Renders the component
     */
    render() {
        if (isNil(this.context) || isNil(this.props.column)) {
            return null;
        }
        const unhandledProps = {
            role: "gridcell",
            tabIndex: this.isFocusCell() ? 0 : -1,
            onKeyDown: this.handleKeyDown,
            onFocus: this.handleFocus,
        };
        const config = {
            rowData: this.props.rowData,
            columnDataKey: this.props.column.columnDataKey,
            columnIndex: this.props.columnIndex,
            classNames: this.generateClassNames(),
            rootElement: this.rootElement,
            focusTarget: this.focusTarget,
            unhandledProps: unhandledProps,
        };
        return !isNil(this.props.column.cell)
            ? this.props.column.cell(config)
            : DataGridCell.renderCell(config);
    }
    /**
     * React life-cycle method
     */
    componentDidMount() {
        if (this.isDesiredFocusCell() && !isNil(this.rootElement.current)) {
            this.rootElement.current.focus();
        }
    }
    /**
     * Generates class names
     */
    generateClassNames() {
        return super.generateClassNames(classNames(this.props.managedClasses.dataGridCell));
    }
}
DataGridCell.defaultProps = {
    rowData: null,
    column: null,
    columnIndex: 0,
    managedClasses: {},
};
DataGridCell.displayName = `${DisplayNamePrefix}DataGridCell`;
DataGridCell.contextType = DataGridContext;
/**
 * The default function that renders a cell
 */
DataGridCell.renderCell = (config) => {
    return (React.createElement("div", Object.assign({}, config.unhandledProps, { ref: config.rootElement, "data-cellid": config.columnDataKey, className: config.classNames, style: {
            gridColumn: config.columnIndex,
        } }), config.rowData[config.columnDataKey]));
};
DataGridCell.contextType = DataGridContext;
export default DataGridCell;
export * from "./data-grid-cell.props";
