import React from "react";
import { isNil } from "lodash-es";
import Foundation from "@microsoft/fast-components-foundation-react";
import { classNames, Direction, KeyCodes } from "@microsoft/fast-web-utilities";
import throttle from "raf-throttle";
import { DisplayNamePrefix } from "../utilities";
import StackPanel from "../stack-panel";
import DataGridRow from "./data-grid-row";
import { DataGridContext } from "./data-grid-context";
class DataGrid extends Foundation {
    /**
     * constructor
     */
    constructor(props) {
        super(props);
        this.handledProps = {
            dataRowKey: void 0,
            rows: void 0,
            columns: void 0,
            rowHeight: void 0,
            rowHeightCallback: void 0,
            managedClasses: void 0,
            defaultFocusColumnKey: void 0,
            defaultFocusRowKey: void 0,
            stableRangeEndIndex: void 0,
            pageSize: void 0,
            virtualize: void 0,
        };
        this.currentTemplateColumns = "";
        this.rootElement = React.createRef();
        this.nonVirtualizedScrollContainer = React.createRef();
        this.direction = Direction.ltr;
        this.isFocused = false;
        this.lastReportedScrollPosition = 0;
        this.lastReportedViewportSpan = 0;
        /**
         * Updates focus row related state after an update
         */
        this.validateFocusRowState = (newState) => {
            const newFocusRowIndex = this.getRowIndexByKey(newState.focusRowKey);
            if (newFocusRowIndex === -1) {
                // our focus row no longer exists, assign a new one based on previous focus row index
                newState.focusRowIndex = Math.min(Math.max(newState.focusRowIndex, 0), this.props.rows.length - 1);
                newState.focusRowKey = this.props.rows[newState.focusRowIndex][this.props.dataRowKey];
            }
            else {
                newState.focusRowIndex = newFocusRowIndex;
            }
            return newState;
        };
        /**
         * Applies changes to focus row and focus column props to the provided state object
         */
        this.applyUpdatedFocusProps = (newState, prevProps) => {
            let shouldUpdateState = false;
            // apply a new focus columnkey if one is provided and it exists
            if (this.props.defaultFocusColumnKey !== prevProps.defaultFocusColumnKey &&
                !isNil(this.props.defaultFocusColumnKey) &&
                this.getColumnIndexByKey(this.props.defaultFocusColumnKey, newState.columns) !== -1) {
                newState.focusColumnKey = this.props.defaultFocusColumnKey;
                shouldUpdateState = true;
            }
            // apply a new focus rowkey if one is provided and it exists
            if (this.props.defaultFocusRowKey !== prevProps.defaultFocusRowKey &&
                !isNil(this.props.defaultFocusRowKey)) {
                const newFocusRowIndex = this.getRowIndexByKey(this.props.defaultFocusRowKey);
                if (newFocusRowIndex !== -1) {
                    newState.focusRowKey = this.props.defaultFocusRowKey;
                    newState.desiredVisibleRowIndex = newFocusRowIndex;
                    newState.focusRowIndex = newFocusRowIndex;
                    newState.currentDataPageStartIndex = Math.max(0, newFocusRowIndex - Math.floor(this.props.pageSize / 2));
                    newState.currentDataPageEndIndex = Math.min(newState.currentDataPageStartIndex + this.props.pageSize, this.props.rows.length - 1);
                    shouldUpdateState = true;
                }
            }
            return shouldUpdateState;
        };
        /**
         * Gets the initial state of the component based on virtualizion prop
         * also used when prop is changed during component lifetime
         */
        this.applyInitialState = () => {
            const newState = this.applyInitialFocusState(this.getInitialStateObject());
            return this.props.virtualize
                ? this.applyVirtualizedInitialState(newState)
                : this.applyNonVirtualizedInitialState(newState);
        };
        /**
         * Gets the virtualized mode initial state
         * also used when prop is changed during component lifetime
         */
        this.applyVirtualizedInitialState = (newState) => {
            this.componentDidUpdate = this.virtualizedComponentUpdateHandler;
            this.renderPanel = this.renderVirtualizingPanel;
            if (this.props.rows.length > 0) {
                if (!isNil(this.props.defaultFocusRowKey)) {
                    newState.focusRowKey = this.props.defaultFocusRowKey;
                    newState.focusRowIndex = this.getRowIndexByKey(newState.focusRowKey);
                }
                if (newState.focusRowIndex === -1) {
                    newState.focusRowIndex = 0;
                    newState.focusRowKey = this.props.rows[0][this.props.dataRowKey];
                }
                newState.currentDataPageStartIndex = Math.max(0, newState.focusRowIndex - Math.floor(this.props.pageSize / 2));
                newState.currentDataPageEndIndex = Math.min(this.props.rows.length - 1, newState.currentDataPageStartIndex + this.props.pageSize);
                this.sizeRowsToIndex(newState.currentDataPageEndIndex, newState.rowPositions);
                newState.estimatedTotalHeight = this.getEstimatedTotalHeight(newState.rowPositions);
            }
            return newState;
        };
        /**
         * Gets the non-virtualized mode initial state
         * also used when prop is changed during component lifetime
         */
        this.applyNonVirtualizedInitialState = (newState) => {
            this.componentDidUpdate = this.nonVirtualizedComponentUpdateHandler;
            this.renderPanel = this.renderNonVirtualizingPanel;
            newState.currentDataPageStartIndex = this.props.rows.length > 0 ? 0 : -1;
            newState.currentDataPageEndIndex =
                this.props.rows.length > 0 ? this.props.rows.length - 1 : -1;
            newState.rowPositions = [];
            newState.estimatedTotalHeight = 0;
            return newState;
        };
        /**
         * sets initial/reset focus rows
         */
        this.applyInitialFocusState = (newState) => {
            if (!isNil(this.props.defaultFocusRowKey)) {
                newState.focusRowKey = this.props.defaultFocusRowKey;
                newState.focusRowIndex = this.getRowIndexByKey(newState.focusRowKey);
            }
            if (newState.focusRowIndex === -1 && this.props.rows.length > 0) {
                newState.focusRowIndex = 0;
                newState.focusRowKey = this.props.rows[0][this.props.dataRowKey];
            }
            if (newState.columns.length > 0) {
                newState.focusColumnKey =
                    !isNil(this.props.defaultFocusColumnKey) &&
                        this.getColumnIndexByKey(this.props.defaultFocusColumnKey, newState.columns) !== -1
                        ? this.props.defaultFocusColumnKey
                        : newState.columns[0].columnDataKey;
            }
            newState.desiredVisibleRowIndex = newState.focusRowIndex;
            return newState;
        };
        /**
         *  render the header
         */
        this.renderGridHeader = () => {
            return (React.createElement("div", { className: this.props.managedClasses.dataGrid_header, role: "row", style: {
                    display: "grid",
                    gridTemplateColumns: this.currentTemplateColumns,
                } }, this.state.columns.map(this.renderColumnHeader)));
        };
        /**
         *  placeholder function until assignement based on virtualization mode
         */
        this.renderPanel = () => {
            return null;
        };
        /**
         *  render a non-virtualizing panel
         */
        this.renderNonVirtualizingPanel = () => {
            return (React.createElement("div", { ref: this.nonVirtualizedScrollContainer, tabIndex: -1, className: this.props.managedClasses.dataGrid_scrollingPanel }, this.renderNonVirtualizedRows()));
        };
        /**
         *  render non virtualized data rows
         */
        this.renderNonVirtualizedRows = () => {
            const rowsToRender = [];
            this.props.rows.forEach((row, index) => {
                rowsToRender.push(this.renderRow(row, index));
            });
            return rowsToRender;
        };
        /**
         *  render a virtualizing panel
         */
        this.renderVirtualizingPanel = () => {
            const itemSpans = [];
            if (this.state.rowPositions.length > 0) {
                itemSpans.push(this.state.rowPositions[this.state.currentDataPageStartIndex].start);
                if (typeof this.props.rowHeightCallback === "function") {
                    for (let i = this.state.currentDataPageStartIndex; i <= this.state.currentDataPageEndIndex; i++) {
                        itemSpans.push(this.state.rowPositions[i].span);
                    }
                }
                itemSpans.push(this.state.estimatedTotalHeight -
                    this.state.rowPositions[this.state.currentDataPageEndIndex].end);
            }
            const stackPanelVisibleItemIndex = this.state.desiredVisibleRowIndex !== null
                ? this.convertRowIndexToStackPanelIndex(this.state.desiredVisibleRowIndex, this.state.currentDataPageStartIndex, this.state.currentDataPageEndIndex)
                : null;
            const { dataGrid_scrollingPanel, dataGrid_scrollingPanelItems, dataGrid_scrollingPanel__scrollable, } = this.props.managedClasses;
            return (React.createElement(StackPanel, { initiallyVisibleItemIndex: stackPanelVisibleItemIndex, onScrollChange: this.throttledScroll, itemSpan: itemSpans, virtualize: this.props.virtualize, tabIndex: -1, style: {
                    height: "100%",
                    overflowY: "scroll",
                }, managedClasses: {
                    stackPanel: dataGrid_scrollingPanel,
                    stackPanel_items: dataGrid_scrollingPanelItems,
                    stackPanel__scrollable: dataGrid_scrollingPanel__scrollable,
                } }, this.renderVirtualizedRows()));
        };
        /**
         *  render the data rows
         */
        this.renderVirtualizedRows = () => {
            const rowsToRender = [];
            rowsToRender.push(React.createElement("div", { key: "frontSpacer" }));
            for (let i = this.state.currentDataPageStartIndex; i <= this.state.currentDataPageEndIndex; i++) {
                rowsToRender.push(this.renderRow(this.props.rows[i], i));
            }
            rowsToRender.push(React.createElement("div", { key: "backSpacer" }));
            return rowsToRender;
        };
        /**
         *  render each column header
         */
        this.renderColumnHeader = (column, index) => {
            const { dataGrid_columnHeader, } = this.props.managedClasses;
            const config = {
                title: isNil(column.title) ? column.columnDataKey : column.title,
                key: column.columnDataKey,
                columnIndex: index,
                classNames: dataGrid_columnHeader,
            };
            if (!isNil(column.header)) {
                return column.header(config);
            }
            else {
                return DataGrid.renderColumnHeader(config);
            }
        };
        /**
         * Render a single data row
         */
        this.renderRow = (rowData, index) => {
            if (isNil(rowData)) {
                return;
            }
            const rowKey = !isNil(rowData[this.props.dataRowKey])
                ? rowData[this.props.dataRowKey]
                : index;
            const { dataGrid_row, dataGrid_row__focusedWithin, dataGrid_cell, } = this.props.managedClasses;
            return (React.createElement(DataGridRow, { key: rowKey, rowIndex: index, rowData: rowData, gridTemplateColumns: this.currentTemplateColumns, managedClasses: {
                    dataGridRow: dataGrid_row,
                    dataGridRow__focusedWithin: dataGrid_row__focusedWithin,
                    dataGridRow_cell: dataGrid_cell,
                } }));
        };
        /**
         *  Generates the grid template column css string
         */
        this.getGridTemplateColumns = () => {
            let templateColumns = "";
            this.state.columns.forEach((column) => {
                templateColumns = `${templateColumns} ${isNil(column.columnWidth) ? "1fr" : column.columnWidth}`;
            });
            return templateColumns;
        };
        /**
         *  Handle grid focus
         */
        this.handleGridFocus = () => {
            this.isFocused = true;
        };
        /**
         *  Handle grid blur by setting focused state
         */
        this.handleGridBlur = (e) => {
            const root = this.rootElement.current;
            // If we focus outside of the data grid
            if (!!root && !root.contains(e.relatedTarget)) {
                this.isFocused = false;
            }
        };
        /**
         * Handle the keydown event of the item
         */
        this.handleCellKeyDown = (cell, e) => {
            if (e.defaultPrevented) {
                return;
            }
            switch (e.keyCode) {
                case KeyCodes.arrowDown:
                    e.preventDefault();
                    this.incrementFocusRow(1);
                    break;
                case KeyCodes.arrowRight:
                    this.incrementFocusColumn(1);
                    e.preventDefault();
                    break;
                case KeyCodes.arrowUp:
                    this.incrementFocusRow(-1);
                    e.preventDefault();
                    break;
                case KeyCodes.arrowLeft:
                    this.incrementFocusColumn(-1);
                    e.preventDefault();
                    break;
                case KeyCodes.pageDown:
                    this.incrementFocusPage(1);
                    e.preventDefault();
                    break;
                case KeyCodes.pageUp:
                    this.incrementFocusPage(-1);
                    e.preventDefault();
                    break;
                case KeyCodes.home:
                    if (e.ctrlKey) {
                        this.incrementFocusRow(-this.props.rows.length);
                    }
                    else {
                        this.incrementFocusColumn(-this.state.columns.length);
                    }
                    e.preventDefault();
                    break;
                case KeyCodes.end:
                    if (e.ctrlKey) {
                        this.incrementFocusRow(this.props.rows.length);
                    }
                    else {
                        this.incrementFocusColumn(this.state.columns.length);
                    }
                    e.preventDefault();
                    break;
            }
        };
        /**
         * get the estimated total height of the datagrid based on row heights calculated so far
         */
        this.getEstimatedTotalHeight = (rowPositions) => {
            if (rowPositions.length === 0) {
                return 0;
            }
            const estimatedTotalHeight = rowPositions[rowPositions.length - 1].end +
                (this.props.rows.length - rowPositions.length) * this.props.rowHeight;
            return estimatedTotalHeight;
        };
        /**
         * calculate total height of rows to target index
         * note: this modifies the provided rowposition array directly
         */
        this.sizeRowsToIndex = (targetIndex, rowPositions) => {
            if (rowPositions.length - 1 < targetIndex) {
                const startIndex = rowPositions.length;
                const endIndex = targetIndex > this.props.rows.length - 1
                    ? this.props.rows.length - 1
                    : targetIndex;
                for (let i = startIndex; i <= endIndex; i++) {
                    const thisRowStart = i === 0 ? 0 : rowPositions[i - 1].end;
                    const thisRowHeight = this.props.rowHeightCallback({
                        rowData: this.props.rows[i],
                        rowIndex: i,
                        defaultRowHeight: this.props.rowHeight,
                    });
                    rowPositions.push({
                        start: thisRowStart,
                        span: thisRowHeight,
                        end: thisRowStart + thisRowHeight,
                    });
                }
            }
        };
        /**
         * size rows to target scroll value
         * appends rowposition data until the bottom of the last item
         * is greater than target scroll value.
         * note: this modifies the provided rowposition array directly
         */
        this.sizeRowsToScrollValue = (targetScrollValue, rowPositions) => {
            if (rowPositions[rowPositions.length - 1].end < targetScrollValue) {
                const startIndex = rowPositions.length;
                const endIndex = this.props.rows.length - 1;
                for (let i = startIndex; i <= endIndex; i++) {
                    const thisRowStart = i === 0 ? 0 : rowPositions[i - 1].end;
                    const thisRowHeight = this.props.rowHeightCallback({
                        rowData: this.props.rows[i],
                        rowIndex: i,
                        defaultRowHeight: this.props.rowHeight,
                    });
                    const thisRowEnd = thisRowStart + thisRowHeight;
                    rowPositions.push({
                        start: thisRowStart,
                        span: thisRowHeight,
                        end: thisRowEnd,
                    });
                    if (thisRowEnd >= targetScrollValue) {
                        break;
                    }
                }
            }
        };
        /**
         * handle scroll events from the stackpanel
         */
        this.handleStackPanelScrollChange = (newScrollValue, scrollMaxValue, viewportSpan) => {
            const currentPageTop = this.state.rowPositions[this.state.currentDataPageStartIndex].start;
            const currentPageBottom = this.state.rowPositions[this.state.currentDataPageEndIndex].end;
            const currentViewportBottom = newScrollValue + viewportSpan;
            this.lastReportedScrollPosition = newScrollValue;
            this.lastReportedViewportSpan = viewportSpan;
            if (this.state.rowPositions[this.state.focusRowIndex].start >
                currentViewportBottom ||
                this.state.rowPositions[this.state.focusRowIndex].end < newScrollValue) {
                const topVisibleElementIndex = this.getIndexOfRowAtScrollPosition(newScrollValue, this.state.rowPositions);
                if (topVisibleElementIndex === -1) {
                    this.setState({
                        focusRowKey: "",
                    });
                }
                else {
                    this.setState({
                        focusRowKey: this.props.rows[topVisibleElementIndex][this.props.dataRowKey],
                    });
                }
            }
            if (newScrollValue < currentPageTop ||
                currentViewportBottom > currentPageBottom) {
                const newRowPositions = this.state.rowPositions.slice(0);
                const middleViewportPosition = Math.floor(newScrollValue + viewportSpan / 2);
                const middleItemIndex = this.getIndexOfRowAtScrollPosition(middleViewportPosition, newRowPositions);
                let newDataPageStartIndex = middleItemIndex - Math.floor(this.props.pageSize / 2);
                if (newDataPageStartIndex < 0) {
                    newDataPageStartIndex = 0;
                }
                let newDataPageEndIndex = newDataPageStartIndex + this.props.pageSize - 1;
                if (newDataPageEndIndex > this.props.rows.length - 1) {
                    newDataPageEndIndex = this.props.rows.length - 1;
                }
                this.sizeRowsToIndex(newDataPageEndIndex, newRowPositions);
                this.setState({
                    currentDataPageStartIndex: newDataPageStartIndex,
                    currentDataPageEndIndex: newDataPageEndIndex,
                    rowPositions: newRowPositions,
                    estimatedTotalHeight: this.getEstimatedTotalHeight(newRowPositions),
                });
            }
        };
        /**
         * returns the index of row item at a particular scroll position
         */
        this.getIndexOfRowAtScrollPosition = (scrollPosition, rowPositions) => {
            // handle virtualizing display
            if (this.props.virtualize) {
                if (rowPositions.length === 0) {
                    return -1;
                }
                scrollPosition = Math.max(0, scrollPosition);
                this.sizeRowsToScrollValue(scrollPosition, rowPositions);
                const maxIndex = Math.min(rowPositions.length - 1);
                const estimatedItemIndex = Math.min(maxIndex, Math.floor(scrollPosition / this.props.rowHeight));
                if (scrollPosition < rowPositions[estimatedItemIndex].start) {
                    for (let i = estimatedItemIndex; i >= 0; i--) {
                        if (rowPositions[i].start < scrollPosition) {
                            return i;
                        }
                    }
                    return 0;
                }
                else if (scrollPosition > rowPositions[estimatedItemIndex].end) {
                    let maxRowpositionSizedIndex = rowPositions.length - 1;
                    for (let i = estimatedItemIndex; i <= maxIndex; i++) {
                        if (i > maxRowpositionSizedIndex) {
                            this.sizeRowsToIndex(i, rowPositions);
                            maxRowpositionSizedIndex++;
                        }
                        if (rowPositions[i].end > scrollPosition) {
                            return i;
                        }
                    }
                    return maxIndex;
                }
                return estimatedItemIndex;
            }
            // handle non-virtualizing display
            if (this.nonVirtualizedScrollContainer.current === null) {
                return;
            }
            const rows = this.getRenderedRows();
            const rowCount = rows.length;
            for (let i = 0; i < rowCount; i++) {
                if (rows[i].offsetTop >= scrollPosition) {
                    return i;
                }
            }
            return rowCount - 1;
        };
        /**
         * move focus up/down one viewport's worth of items
         */
        this.incrementFocusPage = (direction) => {
            const newRowPositions = this.state.rowPositions.slice(0);
            let nextItemIndex = -1;
            if (this.props.virtualize) {
                const nextItemScrollPosition = direction > 0
                    ? this.lastReportedScrollPosition + this.lastReportedViewportSpan + 1
                    : this.lastReportedScrollPosition - this.lastReportedViewportSpan + 1;
                nextItemIndex = this.getIndexOfRowAtScrollPosition(nextItemScrollPosition, newRowPositions);
            }
            else if (this.nonVirtualizedScrollContainer.current !== null) {
                const newScrollPosition = Math.min(direction > 0
                    ? this.nonVirtualizedScrollContainer.current.scrollTop +
                        this.nonVirtualizedScrollContainer.current.clientHeight +
                        1
                    : this.nonVirtualizedScrollContainer.current.scrollTop -
                        this.nonVirtualizedScrollContainer.current.clientHeight +
                        1, this.nonVirtualizedScrollContainer.current.scrollHeight);
                nextItemIndex = this.getIndexOfRowAtScrollPosition(newScrollPosition, this.state.rowPositions);
                this.scrollNonVirtualizedRowToTop(nextItemIndex);
            }
            // don't get stuck on items that span the viewport
            if (nextItemIndex === this.state.focusRowIndex) {
                if (direction > 0 && nextItemIndex < this.props.rows.length - 1) {
                    nextItemIndex++;
                }
                else if (direction < 0 && nextItemIndex > 0) {
                    nextItemIndex;
                }
            }
            this.focusOnCell(this.props.rows[nextItemIndex][this.props.dataRowKey], this.state.focusColumnKey, newRowPositions, true);
        };
        /**
         * move focus to another row
         */
        this.incrementFocusRow = (direction) => {
            let currentFocusRowIndex = this.getRowIndexByKey(this.state.focusRowKey);
            if (currentFocusRowIndex === -1) {
                currentFocusRowIndex =
                    this.state.focusRowIndex < this.props.rows.length
                        ? this.state.focusRowIndex
                        : this.props.rows.length - 1;
            }
            let newFocusRowIndex = currentFocusRowIndex + direction;
            if (newFocusRowIndex < 0) {
                newFocusRowIndex = 0;
            }
            else if (newFocusRowIndex >= this.props.rows.length) {
                newFocusRowIndex = this.props.rows.length - 1;
            }
            const newFocusRowKey = this.props.rows[newFocusRowIndex][this.props.dataRowKey];
            const focusRowElement = this.getRowElementByKey(newFocusRowKey);
            const focusCell = this.getCellElementByKey(this.state.focusColumnKey, focusRowElement);
            if (isNil(focusCell)) {
                return;
            }
            focusCell.focus();
        };
        /**
         *  move focus to another column
         */
        this.incrementFocusColumn = (direction) => {
            this.updateDirection();
            let currentFocusColumnIndex = this.getColumnIndexByKey(this.state.focusColumnKey, this.state.columns);
            if (currentFocusColumnIndex === -1) {
                currentFocusColumnIndex = 0;
            }
            const newFocusColumnIndex = Math.min(Math.max(currentFocusColumnIndex + direction * (Direction.ltr ? 1 : -1), 0), this.state.columns.length - 1);
            const newFocusColumnKey = this.state.columns[newFocusColumnIndex]
                .columnDataKey;
            const rowElement = this.getRowElementByKey(this.state.focusRowKey);
            if (isNil(rowElement)) {
                return;
            }
            const cellElement = this.getCellElementByKey(newFocusColumnKey, rowElement);
            if (cellElement instanceof HTMLElement) {
                cellElement.focus();
            }
        };
        /**
         *  Get all rendered rows
         */
        this.getRenderedRows = () => {
            if (isNil(this.rootElement.current)) {
                return null;
            }
            let rows = Array.from(this.rootElement.current.querySelectorAll("[role='row']"));
            // remove the header row
            if (rows.length > 0) {
                rows = rows.slice(1);
            }
            return rows;
        };
        /**
         *  Get row element by key
         */
        this.getRowElementByKey = (rowId) => {
            if (isNil(this.rootElement.current)) {
                return null;
            }
            return this.rootElement.current.querySelector(`[data-rowid=${rowId}]`);
        };
        /**
         *  Get cell element by key
         */
        this.getCellElementByKey = (columnKey, rowElement) => {
            if (isNil(rowElement)) {
                return null;
            }
            return rowElement.querySelector(`[data-cellid=${columnKey}]`);
        };
        /**
         *  Move focus to a cell in the whole dataset based on row and cell id
         */
        this.focusOnCell = (rowId, cellId, rowPositions, forceScrollRowToTop) => {
            const rowIndex = this.getRowIndexByKey(rowId);
            if (rowIndex === -1) {
                //no such row id
                return;
            }
            if (rowIndex >= this.state.currentDataPageStartIndex &&
                rowIndex <= this.state.currentDataPageEndIndex) {
                // see if the element is in the dom and focus it, otherwise falls back to setting desired focus
                const rowElement = this.getRowElementByKey(rowId);
                if (rowElement !== null) {
                    const cellElement = this.getCellElementByKey(cellId, rowElement);
                    this.setState({
                        desiredVisibleRowIndex: forceScrollRowToTop ? rowIndex : null,
                    });
                    if (cellElement !== null) {
                        cellElement.focus();
                        return;
                    }
                }
            }
            else {
                // shift the items passed to the stack panel to include the desired row indexes
                let newDataPageStartIndex = rowIndex - Math.floor(this.props.pageSize / 2);
                if (newDataPageStartIndex < 0) {
                    newDataPageStartIndex = 0;
                }
                let newDataPageEndIndex = newDataPageStartIndex + this.props.pageSize - 1;
                if (newDataPageEndIndex > this.props.rows.length - 1) {
                    newDataPageEndIndex = this.props.rows.length - 1;
                }
                this.sizeRowsToIndex(newDataPageEndIndex, rowPositions);
                this.setState({
                    currentDataPageStartIndex: newDataPageStartIndex,
                    currentDataPageEndIndex: newDataPageEndIndex,
                    rowPositions: rowPositions,
                    estimatedTotalHeight: this.getEstimatedTotalHeight(rowPositions),
                });
            }
            this.setState({
                desiredVisibleRowIndex: forceScrollRowToTop ? rowIndex : null,
                desiredFocusRowKey: rowId,
                desiredFocusColumnKey: cellId,
            });
        };
        /**
         *  Get column index by key
         */
        this.getColumnIndexByKey = (columnKey, columns) => {
            return columns.findIndex((column) => {
                return column.columnDataKey === columnKey;
            });
        };
        /**
         * Get row index by key
         */
        this.getRowIndexByKey = (rowKey) => {
            const rowIndex = this.props.rows.findIndex((dataRow) => {
                return dataRow[this.props.dataRowKey] === rowKey;
            });
            return rowIndex;
        };
        /**
         * Handle focus event
         */
        this.handleCellFocus = (cell, e) => {
            if (e.defaultPrevented) {
                return;
            }
            const focusRowKey = cell.rowData[this.props.dataRowKey];
            this.setState({
                focusRowKey,
                focusColumnKey: cell.column.columnDataKey,
                focusRowIndex: this.getRowIndexByKey(focusRowKey),
                desiredFocusRowKey: null,
                desiredFocusColumnKey: null,
            });
        };
        /**
         *  Updates the direction
         */
        this.updateDirection = () => {
            if (this.rootElement.current === null) {
                return;
            }
            const closest = this.rootElement.current.closest(`[dir]`);
            if (closest === null) {
                this.direction = Direction.ltr;
                return;
            }
            this.direction =
                closest.getAttribute("dir").toLowerCase() === "rtl"
                    ? Direction.rtl
                    : Direction.ltr;
        };
        /**
         * Converts a row index in the base dataset to an index in the current data page
         * passed to the stack panel.  Returns -1 if outside that range.
         */
        this.convertRowIndexToStackPanelIndex = (rowIndex, dataPageStartIndex, dataPageEndIndex) => {
            if (rowIndex < dataPageStartIndex || rowIndex > dataPageEndIndex) {
                return -1;
            }
            return rowIndex - dataPageStartIndex + 1;
        };
        /**
         * Scrolls the desired row into view in non-virtualized mode
         */
        this.scrollNonVirtualizedRowToTop = (rowIndex) => {
            if (!this.props.virtualize &&
                !isNil(this.nonVirtualizedScrollContainer.current) &&
                this.props.rows.length > rowIndex &&
                rowIndex > -1) {
                const rows = this.getRenderedRows();
                if (rows.length <= rowIndex) {
                    return;
                }
                this.nonVirtualizedScrollContainer.current.scrollTop =
                    rows[rowIndex].offsetTop;
            }
        };
        /**
         *  gets a state object with initial values
         */
        this.getInitialStateObject = () => {
            const newState = {
                focusRowIndex: -1,
                focusRowKey: null,
                focusColumnKey: null,
                columns: this.getColumns(),
                currentDataPageStartIndex: -1,
                currentDataPageEndIndex: -1,
                rowPositions: [],
                estimatedTotalHeight: 0,
                desiredVisibleRowIndex: null,
                desiredFocusRowKey: null,
                desiredFocusColumnKey: null,
            };
            return newState;
        };
        /**
         *  gets the current column configuration
         */
        this.getColumns = () => {
            return isNil(this.props.columns)
                ? this.props.rows.length > 0
                    ? DataGrid.generateColumns(this.props.rows[0])
                    : null
                : this.props.columns;
        };
        this.throttledScroll = throttle(this.handleStackPanelScrollChange);
        this.state = this.applyInitialState();
    }
    /**
     * Renders the component
     */
    render() {
        this.currentTemplateColumns = this.getGridTemplateColumns();
        return (React.createElement(DataGridContext.Provider, { value: {
                onCellFocused: this.handleCellFocus,
                onCellKeyDown: this.handleCellKeyDown,
                focusRowKey: this.state.focusRowKey,
                focusColumnKey: this.state.focusColumnKey,
                desiredFocusColumnKey: this.state.desiredFocusColumnKey,
                desiredFocusRowKey: this.state.desiredFocusRowKey,
                desiredVisibleRowIndex: this.state.desiredVisibleRowIndex,
                columns: this.state.columns,
                dataRowKey: this.props.dataRowKey,
            } },
            React.createElement("div", Object.assign({}, this.unhandledProps(), { className: this.generateClassNames(), role: "grid", tabIndex: -1, onFocus: this.handleGridFocus, onBlur: this.handleGridBlur, ref: this.rootElement }),
                this.renderGridHeader(),
                this.renderPanel())));
    }
    /**
     * Generates class names
     */
    generateClassNames() {
        const { dataGrid, dataGrid__virtualized, } = this.props.managedClasses;
        return super.generateClassNames(classNames(dataGrid, [dataGrid__virtualized, this.props.virtualize]));
    }
    /**
     * React life-cycle method
     */
    componentDidMount() {
        if (this.state.desiredVisibleRowIndex !== null) {
            this.scrollNonVirtualizedRowToTop(this.state.desiredVisibleRowIndex);
        }
        this.setState({
            desiredVisibleRowIndex: null,
        });
    }
    /**
     * componentDidUpdate when in non-virtualized mode
     */
    nonVirtualizedComponentUpdateHandler(prevProps) {
        if (this.props.virtualize) {
            // virtualization mode changed, reset
            this.setState(this.applyInitialState());
            return;
        }
        let shouldUpdateState = false;
        const newState = Object.assign({}, this.state);
        shouldUpdateState = this.applyUpdatedFocusProps(newState, prevProps);
        if (this.props.columns !== prevProps.columns) {
            newState.columns = this.getColumns();
            shouldUpdateState = true;
        }
        if (this.state.desiredVisibleRowIndex !== null) {
            // move desired row into view and note that a state update is required
            this.scrollNonVirtualizedRowToTop(this.state.desiredVisibleRowIndex);
            newState.desiredVisibleRowIndex = null;
            shouldUpdateState = true;
        }
        // revalidate when there is new data
        if (this.props.rows !== prevProps.rows) {
            shouldUpdateState = true;
            if (this.props.rows.length === 0) {
                this.getInitialStateObject();
            }
            else {
                // ensure focus is still valid
                this.validateFocusRowState(newState);
                // ensure data page conforms to new data length
                // (not virtualizing, so cover the whole range)
                newState.currentDataPageStartIndex = 0;
                newState.currentDataPageEndIndex = this.props.rows.length - 1;
                // move focus to the new element if necessary
                if ((this.isFocused && this.state.focusRowKey !== newState.focusRowKey) ||
                    this.state.focusColumnKey !== newState.focusColumnKey) {
                    const rowElement = this.getRowElementByKey(newState.focusRowKey);
                    if (rowElement !== null) {
                        rowElement.focus();
                    }
                }
            }
        }
        if (shouldUpdateState) {
            this.setState(newState);
        }
    }
    /**
     * componentDidUpdate when in virtualized mode
     */
    virtualizedComponentUpdateHandler(prevProps) {
        if (!this.props.virtualize) {
            // virtualization mode changed, reset
            this.setState(this.applyInitialState());
            return;
        }
        let shouldUpdateState = false;
        let newState = Object.assign({}, this.state);
        newState.desiredVisibleRowIndex = null;
        if (this.state.desiredVisibleRowIndex !== null) {
            // note that a state update is required
            shouldUpdateState = true;
        }
        shouldUpdateState = this.applyUpdatedFocusProps(newState, prevProps);
        if (this.props.columns !== prevProps.columns) {
            newState.columns = this.getColumns();
            shouldUpdateState = true;
        }
        // revalidate when there is new data
        if (this.props.rows !== prevProps.rows) {
            shouldUpdateState = true;
            if (this.props.rows.length === 0) {
                newState = this.getInitialStateObject();
            }
            else {
                // if an author tells us nothing has changed before a certain point we keep position data
                // up to that point.
                newState.rowPositions = newState.rowPositions.slice(0, this.props.stableRangeEndIndex);
                // ensure focus is still valid
                this.validateFocusRowState(newState);
                // ensure data page conforms to new data length
                newState.currentDataPageStartIndex =
                    newState.focusRowIndex - Math.floor(this.props.pageSize / 2);
                newState.currentDataPageStartIndex = Math.max(0, newState.currentDataPageStartIndex);
                newState.currentDataPageEndIndex =
                    newState.currentDataPageStartIndex + this.props.pageSize;
                newState.currentDataPageEndIndex =
                    newState.currentDataPageEndIndex > this.props.rows.length - 1
                        ? this.props.rows.length - 1
                        : newState.currentDataPageEndIndex;
                this.sizeRowsToIndex(newState.currentDataPageEndIndex, newState.rowPositions);
                newState.estimatedTotalHeight = this.getEstimatedTotalHeight(newState.rowPositions);
                // move focus to the new element if necessary
                if ((this.isFocused && this.state.focusRowKey !== newState.focusRowKey) ||
                    this.state.focusColumnKey !== newState.focusColumnKey) {
                    const rowElement = this.getRowElementByKey(newState.focusRowKey);
                    if (rowElement !== null) {
                        // next focus element already exists, focus on it
                        rowElement.focus();
                        newState.desiredFocusRowKey = null;
                        newState.desiredFocusColumnKey = null;
                    }
                    else {
                        // next focus element not in the dom,
                        // focus on it after when it mounts
                        newState.desiredFocusRowKey = newState.focusRowKey;
                        newState.desiredFocusColumnKey = newState.focusColumnKey;
                        // let the default behavior scroll into view on focus
                        newState.desiredVisibleRowIndex = null;
                    }
                }
            }
        }
        if (shouldUpdateState) {
            this.setState(newState);
        }
    }
}
DataGrid.defaultProps = {
    rowHeight: 60,
    stableRangeEndIndex: 0,
    pageSize: 1000,
    rowHeightCallback: (row) => {
        return row.defaultRowHeight;
    },
    virtualize: false,
    managedClasses: {},
};
DataGrid.displayName = `${DisplayNamePrefix}DataGrid`;
/**
 *  default column render function
 */
DataGrid.renderColumnHeader = (config) => {
    return (React.createElement("div", { className: config.classNames, role: "columnheader", key: config.key, style: {
            gridColumn: config.columnIndex + 1,
        } }, config.title));
};
/**
 *  generates a basic column definition by examining sample row data
 */
DataGrid.generateColumns = (row) => {
    const definitions = [];
    const properties = Object.getOwnPropertyNames(row);
    properties.forEach((property) => {
        definitions.push({
            columnDataKey: property,
        });
    });
    return definitions;
};
export default DataGrid;
export * from "./data-grid.props";
export { DataGridContext };
