import Foundation from "@microsoft/fast-components-foundation-react";
import { classNames, Direction, RtlScrollConverter } from "@microsoft/fast-web-utilities";
import { get, isFunction, isNil, merge } from "lodash-es";
import React from "react";
import { toPx } from "@microsoft/fast-jss-utilities";
import { Orientation } from "@microsoft/fast-web-utilities";
import { DisplayNamePrefix } from "../utilities";
class StackPanel extends Foundation {
    /**
     * constructor
     */
    constructor(props) {
        super(props);
        this.handledProps = {
            itemSpan: void 0,
            virtualize: void 0,
            neverVirtualizeIndexes: void 0,
            preloadBufferCount: void 0,
            orientation: void 0,
            managedClasses: void 0,
            initiallyVisibleItemIndex: void 0,
            onScrollChange: void 0,
            scrollLayoutUpdateDelay: void 0,
        };
        // ref to the root element of the component
        this.rootElement = React.createRef();
        /**
         * ref to the element which is the container all children are rendered into
         * It is sized accomodate all children and is the container that is being scrolled
         */
        this.itemContainerElement = React.createRef();
        this.viewportSpan = 0;
        this.itemContainerSpan = 0;
        this.maxScroll = 0;
        /**
         * Stores last scroll position from scroll events
         */
        this.lastRecordedScroll = 0;
        /**
         * Delays updating ui during scrolling
         * (to avoid rendering of items that just scroll by)
         */
        this.scrollLayoutUpdateTimer = null;
        /**
         * Render a single item
         */
        this.renderItem = (item, index) => {
            if (index >= this.itemPositions.length ||
                (this.props.virtualize &&
                    !this.props.neverVirtualizeIndexes.includes(index) &&
                    (index < this.state.renderedRangeStartIndex ||
                        index > this.state.renderedRangeEndIndex))) {
                return;
            }
            const newStyleProps = Object.assign({ position: "absolute" }, (this.props.orientation === Orientation.horizontal
                ? Object.assign({ width: toPx(this.itemPositions[index].span) }, (this.state.direction === Direction.ltr
                    ? { left: toPx(this.itemPositions[index].start) }
                    : { right: toPx(this.itemPositions[index].start) })) : {
                height: toPx(this.itemPositions[index].span),
                top: toPx(this.itemPositions[index].start),
            }));
            if (React.isValidElement(item)) {
                const styleProps = get(item.props, "style");
                return React.cloneElement(item, {
                    style: merge({}, styleProps, newStyleProps),
                });
            }
            // it's a string, so wrap it in a div for positioning
            return React.createElement("div", { style: newStyleProps }, item);
        };
        /**
         * Builds a list of positioning data for all items
         */
        this.updateItemPositions = () => {
            const itemPositions = [];
            let currentItemStart = 0;
            const childrenAsArray = React.Children.toArray(this.props.children);
            const itemCount = childrenAsArray.length;
            for (let i = 0; i < itemCount; i++) {
                const itemSpan = Array.isArray(this.props.itemSpan)
                    ? this.props.itemSpan[i]
                    : this.props.itemSpan;
                const currentItemBottom = currentItemStart + itemSpan;
                itemPositions.push({
                    span: itemSpan,
                    start: currentItemStart,
                    end: currentItemBottom,
                });
                currentItemStart = currentItemBottom;
            }
            this.itemPositions = itemPositions;
            this.itemContainerSpan =
                this.itemPositions.length > 0 ? this.itemPositions[itemCount - 1].end : 0;
            this.maxScroll = this.getMaxScrollDistance();
        };
        /**
         *  Get the size of the viewport along the panel's axis
         */
        this.getViewportSpan = () => {
            if (isNil(this.rootElement.current)) {
                return 0;
            }
            if (this.props.orientation === Orientation.horizontal) {
                return this.rootElement.current.clientWidth;
            }
            else {
                return this.rootElement.current.clientHeight;
            }
        };
        /**
         * Recalculates the layout
         */
        this.updateLayout = () => {
            if (isNil(this.rootElement.current) || isNil(this.itemContainerElement.current)) {
                return;
            }
            if (this.props.orientation === Orientation.horizontal) {
                this.updateDirection();
            }
            const scrollPos = this.getScrollPosition();
            const lastIndex = this.itemPositions.length === 0 ? 0 : this.itemPositions.length - 1;
            const visibleRangeStartIndex = this.getThresholdItemIndex(this.itemPositions, 0, scrollPos);
            const visibleRangeEndIndex = this.getThresholdItemIndex(this.itemPositions, visibleRangeStartIndex, scrollPos + this.viewportSpan);
            const bufferLength = this.getBufferLength();
            window.clearTimeout(this.scrollLayoutUpdateTimer);
            this.setState({
                renderedRangeStartIndex: Math.max(visibleRangeStartIndex - bufferLength, 0),
                renderedRangeEndIndex: Math.min(visibleRangeEndIndex + bufferLength, lastIndex),
                isScrollable: this.itemContainerSpan > this.viewportSpan,
            });
        };
        /**
         * gets the first row where the row bottom exceeds the threshold value
         * returns final item index if no rows reach threshold value
         */
        this.getThresholdItemIndex = (itemPositions, startIndex, threshold) => {
            if (itemPositions.length === 0) {
                return 0;
            }
            let thresholdIndex = itemPositions.length - 1;
            for (let i = startIndex, rowCount = itemPositions.length; i < rowCount; i++) {
                const thisItemPosition = itemPositions[i];
                if (thisItemPosition.end >= threshold) {
                    thresholdIndex = i;
                    break;
                }
            }
            return thresholdIndex;
        };
        /**
         * gets the current buffer size (number of items to render outside the viewport)
         */
        this.getBufferLength = () => {
            if (!isNil(this.props.preloadBufferCount) && this.props.preloadBufferCount >= 0) {
                return Math.floor(this.props.preloadBufferCount);
            }
            return 1;
        };
        /**
         *  Handle scroll events
         */
        this.onScrollCapture = () => {
            this.lastRecordedScroll = this.getScrollPosition();
            clearTimeout(this.scrollLayoutUpdateTimer);
            if (this.props.scrollLayoutUpdateDelay !== 0) {
                this.scrollLayoutUpdateTimer = window.setTimeout(() => {
                    this.updateLayout();
                }, this.props.scrollLayoutUpdateDelay);
            }
            else {
                this.updateLayout();
            }
            this.reportScrollChange();
        };
        /**
         *  Handle viewport resize events
         */
        this.handleResize = () => {
            this.viewportSpan = this.getViewportSpan();
            this.reportScrollChange();
            this.updateLayout();
        };
        /**
         *  report scroll changes to callback
         */
        this.reportScrollChange = () => {
            if (isFunction(this.props.onScrollChange)) {
                this.props.onScrollChange(this.lastRecordedScroll, this.maxScroll, this.viewportSpan);
            }
        };
        /**
         *  updates the direction in state if necessary
         */
        this.updateDirection = () => {
            const newDirection = this.getDirection();
            if (newDirection !== this.state.direction) {
                this.setState({
                    direction: newDirection,
                });
            }
        };
        /**
         *  gets the current direction
         */
        this.getDirection = () => {
            if (this.rootElement.current === null) {
                return Direction.ltr;
            }
            const closest = this.rootElement.current.closest(`[${StackPanel.DirectionAttributeName}]`);
            return closest === null ||
                closest.getAttribute(StackPanel.DirectionAttributeName) === Direction.ltr
                ? Direction.ltr
                : Direction.rtl;
        };
        /**
         *  Gets the scroll position and accounts for direction
         */
        this.getScrollPosition = () => {
            if (isNil(this.rootElement.current)) {
                return 0;
            }
            return this.props.orientation === Orientation.vertical
                ? this.rootElement.current.scrollTop
                : this.state.direction === Direction.rtl
                    ? RtlScrollConverter.getScrollLeft(this.rootElement.current, this.state.direction)
                    : -RtlScrollConverter.getScrollLeft(this.rootElement.current, this.state.direction);
        };
        /**
         *  Sets the scroll position and accounts for direction
         */
        this.setScrollPosition = (scrollValue) => {
            if (isNil(this.rootElement.current)) {
                return;
            }
            if (this.props.orientation === Orientation.vertical) {
                this.rootElement.current.scrollTop = scrollValue;
            }
            else {
                RtlScrollConverter.setScrollLeft(this.rootElement.current, this.state.direction === Direction.rtl ? -scrollValue : scrollValue, this.state.direction);
            }
        };
        /**
         * Gets the scroll position that brings the item at a particular index into the viewport
         */
        this.getScrollIntoViewPosition = (index) => {
            if (index === -1 || index >= this.itemPositions.length) {
                return 0;
            }
            const itemPosition = this.itemPositions[index];
            return Math.max(0, Math.min(itemPosition.start, this.maxScroll));
        };
        this.updateItemPositions();
        this.state = {
            renderedRangeStartIndex: 0,
            renderedRangeEndIndex: 0,
            direction: Direction.ltr,
            isScrollable: false,
        };
    }
    /**
     * Renders the component
     */
    render() {
        const { stackPanel_items, } = this.props.managedClasses;
        return (React.createElement("div", Object.assign({}, this.unhandledProps(), { ref: this.rootElement, className: this.generateClassNames(), onScrollCapture: this.onScrollCapture, tabIndex: -1 }),
            React.createElement("div", { className: classNames(stackPanel_items), ref: this.itemContainerElement, style: Object.assign({ position: "relative" }, (this.props.orientation === Orientation.horizontal
                    ? {
                        height: "100%",
                        width: `${this.itemContainerSpan}px`,
                        overflow: "hidden",
                    }
                    : {
                        width: "100%",
                        height: `${this.itemContainerSpan}px`,
                        overflow: "hidden",
                    })) }, this.renderItems())));
    }
    /**
     * React life-cycle method
     */
    componentDidMount() {
        // set up resize observer so we can update layout when component is resized
        if (!isNil(this.rootElement.current)) {
            if (window.ResizeObserver) {
                this.resizeDetector = new window.ResizeObserver(this.handleResize);
                this.resizeDetector.observe(this.rootElement.current);
            }
        }
        this.viewportSpan = this.getViewportSpan();
        this.updateLayout();
        // if we need to have a particular child scrolled into view on mount do so
        if (!isNil(this.props.initiallyVisibleItemIndex) &&
            this.itemPositions.length > this.props.initiallyVisibleItemIndex) {
            const newScrollPosition = this.getScrollIntoViewPosition(this.props.initiallyVisibleItemIndex);
            this.setScrollPosition(newScrollPosition);
        }
    }
    /**
     * React life-cycle method
     */
    componentDidUpdate(prevProps) {
        if (prevProps.orientation !== this.props.orientation) {
            this.viewportSpan = this.getViewportSpan();
        }
        if (prevProps.itemSpan !== this.props.itemSpan ||
            prevProps.orientation !== this.props.orientation ||
            prevProps.children !== this.props.children) {
            this.updateItemPositions();
            this.updateLayout();
        }
        // if the initiallyVisibleItemIndex prop is changed to a non-null value
        // after initial mount we scroll to that position
        if (prevProps.initiallyVisibleItemIndex !==
            this.props.initiallyVisibleItemIndex &&
            !isNil(this.props.initiallyVisibleItemIndex) &&
            this.itemPositions.length > this.props.initiallyVisibleItemIndex) {
            const newScrollPosition = this.getScrollIntoViewPosition(this.props.initiallyVisibleItemIndex);
            this.scrollContent(newScrollPosition);
        }
    }
    /**
     * React life-cycle method
     */
    componentWillUnmount() {
        if (!isNil(this.rootElement.current)) {
            if (window.ResizeObserver) {
                this.resizeDetector.disconnect();
                this.resizeDetector = null;
            }
        }
        window.clearTimeout(this.scrollLayoutUpdateTimer);
    }
    /**
     * Create class-names
     */
    generateClassNames() {
        const { stackPanel, stackPanel__scrollable, } = this.props.managedClasses;
        return super.generateClassNames(classNames(stackPanel, [stackPanel__scrollable, this.state.isScrollable]));
    }
    /**
     * Render the items
     */
    renderItems() {
        return React.Children.map(this.props.children, this.renderItem);
    }
    /**
     * Gets the maximum distance that can be scrolled
     */
    getMaxScrollDistance() {
        if (this.itemPositions.length === 0) {
            return 0;
        }
        const contentSpan = this.itemPositions[this.itemPositions.length - 1].end;
        return contentSpan - this.viewportSpan;
    }
    /**
     * Scrolls the container for the items list
     */
    scrollContent(targetScrollPosition) {
        const newScrollPosition = Math.max(0, Math.min(targetScrollPosition, this.maxScroll));
        this.setScrollPosition(newScrollPosition);
        this.updateLayout();
    }
}
StackPanel.displayName = `${DisplayNamePrefix}StackPanel`;
StackPanel.defaultProps = {
    managedClasses: {},
    virtualize: true,
    neverVirtualizeIndexes: [],
    itemSpan: 100,
    preloadBufferCount: 1,
    orientation: Orientation.vertical,
    scrollLayoutUpdateDelay: 0,
};
StackPanel.DirectionAttributeName = "dir";
export default StackPanel;
export * from "./stack-panel.props";
